/*======================================================================*/
/*  TITLE:   PLF                                                        */
/*  Function:  Playfield Routines                                       */
/*                                                                      */
/*  First Edit:  10/19/88                                               */
/*  Project #:  532xx                                                   */
/*  Programmer:  Dennis Harper                                          */
/*                                                                      */
/*     COPYRIGHT 1990/1991 ATARI GAMES CORP.                            */
/*     UNATHORIZED REPRODUCTION, ADAPTATION, DISTRIBUTION,              */
/*     PERFORMANCE OR DISPLAY OF THIS COMPUTER PROGRAM OR               */
/*     THE ASSOCIATED AUDIOVISUAL WORK IS STRICTLY PROHIBITED.          */
/*             ALL RIGHTS RESERVED.                                     */
/*                                                                      */
/*----------------------------------------------------------------------*/

#include	"rrdef.h"

#define	SCROLL_CENTER	0xAC0
#define	BOTTOM_SCRL	0x7FC0
#define	SEG_WIDTH	0x1000
#define	SEG_CNT		12

#define	PLF_VSIZE	(0x4000 * 0x40)
#define	BG_VSIZE	0x4000

extern	long	viewx,viewy;
extern	ulong	bckdsp;
extern	WORD	ttlpal[],TTL000[];
extern	WORD	world,viewz_old,strip_num;
extern	WORD	*bkg_tbl[],*bkgx_tbl[],*road_tbl[],*dirt_tbl[],*pfpals[];
extern   WORD  *trklut[];  
extern	WORD	*altstrp[];
extern	WORD	pfram[],pframx[],splpal_tbl[];
extern	short	*rrtracks[],*slopes[];
/* extern   short *alts[]; */
extern	short	curve;
extern	short	viewz,viewy_spd,viewz_off,height;
extern	short	recip[],lacc_tbl[];
extern	BYTE	bank_tbl[];
extern	BYTE	edtflg,lut_mask[];
extern   int   h_scroll[]; /* used to hold road scroll values */
/* extern	BYTE	old_style; */
   BYTE	old_style; 

typedef	struct	scnlin
	  {
	  short	lineh;			/* Hscroll values for each line	*/
	  short	linev;			/* Vscroll values for each line	*/
	  short	lcurve;			/* Curve offset for each line (holding buffer)	*/
	  } SCNLIN;

	SCNLIN	scan_data[SCANLINES];	/* Scan line frame buffer 0	*/
	SCNLIN	*scan_ptr;		/* Pointers for above		*/
	SCNLIN	*hi_scan;		/* Save of highest pointer	*/
	DBL	xshift,ldelta;		/* X shifting values		*/
	DBL	curve_acc;		/* Curve accumulator		*/
	ulong	curve_abs;		/* Curve shift absoulute value 	*/
	PFSCRL	*horizon_ptr;		/* Pointer to horizon scroll	*/
	int	work_line;		/* Current scanline index	*/
	long	curve_chk;		/* Curve boundary check		*/
	long	curve_val;		/* Curve shifting value		*/
	long	max_cacc;		/* Max curve accel		*/
	long	linepos;		/* Current line position	*/
	long	viewy_bb;		/* Backround scroll boundary	*/
	long	maskx;			/* Current mask column for pf	*/
	long	bottomy;		/* Bottom of screen y pos	*/
	long	vscan_seg[SEG_CNT+1];	/* Segment tables		*/
	long	vscrl_seg[SEG_CNT+1];
	WORD	vscan_pri[SEG_CNT+1];
	WORD	seg_pri[SEG_CNT+1];
	WORD	bottom_scroll;		/* Bottom vscroll value		*/
	WORD	vpoint;			/* Vanishing point		*/
	WORD	lut;			/* Current PF LUT		*/
	WORD	trkidx;			/* Track index			*/
	WORD	trk_idx;		/* Track table index		*/
	WORD	wld_edge;		/* Edge of world boundary	*/
	WORD	z_top;			/* Top of world boundary	*/
	WORD	zero_slope;		/* Zero Slope constant		*/
	WORD	stripy;			/* Y delta into current strip	*/
	WORD	bkg_hscroll;		/* Background scroll line shadow	*/
	WORD	seg_bottom,slope_norm;
	WORD	shift_scale;
	WORD	alt_strip;						/* Alternate strip value */
   WORD	horizon;
	short	slope_tbl[MAX_STRIP];	/* Slope table			*/
	short	horizon_norm;		/* Normal horizon value		*/
	short	horizon_line;		/* Current horizon scanline	*/
	short	horizon_old;		/* Last horizon scroll line	*/
	short	line_total;		/* Total of different scanlines	*/
	short	bg_bottom;		/* Bottom of scrolling background	*/
	short	trkang;			/* Track angle, signed		*/
	short	lines;			/* Current working number of scanlines	*/
	short	extra_lines;		/* Extra scan line count	*/
	short	scan_pri;		/* Scanline priority		*/
	short	slope;			/* Current zslope for vscroll	*/
	short	left_bank;		/* Left River bank boundary	*/
	short	right_bank;		/* Right River bank boundary	*/
	short	turn_cur;		/* Current turn			*/
	short	dvz;			/* Delta viewz from playfield	*/
	short	bottom_dy;		/* Deltay of bottom line	*/
	short	*trkptr;		/* Pointer to current track table */
	short	*slope_ptr;		/* Pointer to current slope table */
	short	*altptr;		/* Alternate strip table		*/
	WORD	*lutptr;		/* Pointer to current PF LUT spec table	*/
	BYTE	seg;			/* Current segment index		*/
	BYTE	bg_bank;		/* Current background playfield bank	*/
	BYTE	doing_pf;		/* PF making flag		*/
	BYTE	first_line;		/* First line flag		*/
	BYTE	pf_scroll;		/* PF scroll enable		*/

   WORD	stamp_line;


   /*******************************/


/*   Return deltay from scanline position         */
get_dy(line,dz)
WORD   line;
short   dz;
{
/*   return((0xC000 * viewz)/(line - vpoint));   */
   return((0xC000 * dz)/(line - horizon_norm));
}

/*   Return scanline from deltay      */
get_vscan(dy,dz)
long   dy;
short   dz;
{
                        /* Find vpos of line at deltay   */
   return(vpoint + ((recip[(dy-FRONT_CLIP) >> 2] * dz) / 0x155) & 0xFFC0); 
}


/*   Return scanline from deltay      */
get_vnorm(dy)
long   dy;
{                        /* Find normal vpos of line at deltay   */
   return(((horizon_norm + ((0xC000 * Z_LOWEST)/dy)) + 0x20) & 0xFFC0);
}


/*   Return vscroll line from scanline      */
get_vscroll(line)
WORD   line;
{
   WORD   lines;

   lines = (seg_bottom - line) >> 6;            /* Get number of total number of lines   */
   return(((BOTTOM_SCRL - (lines * slope_norm)) + 0x20) & 0xFFC0);   /* Return vscroll value at that line   */
}



/*   Return height from ypos            */
get_height(ypos)
long   ypos;
{
   WORD   trk_nxt;
   short   partial;

   if ((trk_idx = (WORD)(ypos >> 11)) >= strip_num) trk_idx -= strip_num;   /* Set track index (wrap)   */
   if ((trk_nxt = trk_idx + 1) >= strip_num) trk_nxt -= strip_num;      /* Set next track index (wrap)   */
   partial = ypos & (STRIP_SIZE-1);         /* Get partial strip         */

#if   SLOOP_on
   trk_page();                  /* Set track SLOOP page         */
#endif

   if (old_style)                  /* Get delta viewz from slope tables   */
      return((slope_ptr[trk_idx] + (((slope_ptr[trk_nxt] - slope_ptr[trk_idx]) * partial) >> 11)) >> 1);
   else
      return(slope_ptr[trk_idx] + (((slope_ptr[trk_nxt] - slope_ptr[trk_idx]) * partial) >> 11));
}

/*   Returns pointer to scroll register pair by scanline index      */
PFSCRL   *get_scroll(line)
int   line;
{
/*   return((PFSCRL*)(SCROLL + ((((line & 0xFFFFFFF8) << 2) + (line & 7)) << 2)));*/
}

/**********************************************/


/*********************************/
/*   Init a test playfield         */
/*********************************/
pfinit(wld)
int   wld;
{
   int   i,j;

   trkptr = rrtracks[wld];               /* Track pointer      */
   slope_ptr = slopes[wld];            /* Slope pointer      */
   lutptr = trklut[wld];                /* LUT pointer         */
   old_style = 1;
#if   0
   draw_bg(wld);            /* Draw background   */
   draw_road(wld);            /* Draw road      */
#endif
   init_scroll();            /* Init scroller   */

#if   0
   if (!edtflg)
      draw_dash();            /* Draw dashboard   */

   set_map();            /* Setup track map   */

#if   SLOOP_ON
   trksum();            /* Hacker check      */
#endif
#endif

   trk_idx = 0;            /* Clear track index      */
   maskx = 0;            /* Default middle screen   */
   doing_pf = FALSE;
   left_bank = -WLD_EDGE;         /* Set river bank boundary   */
   right_bank = WLD_EDGE;
   wld_edge = WLD_EDGE;         /* Init world edge      */
   z_top = Z_TOP;            /* Init world top      */

#if   SLOOP_ON
   trk_page();            /* Get track SLOOP page      */
#endif

   for (i=0; i < strip_num; i++)      /* Fill slope table      */
   {
      if ((j = i + 1) >= strip_num)      /* Get slope +1         */
       j -= strip_num;
     
      if (old_style)
           slope_tbl[i] = (slope_ptr[j] - slope_ptr[i]) >> 1;
      else
           slope_tbl[i] = slope_ptr[j] - slope_ptr[i];
   }
}

/*********************************/
/*   Init scroll registers           */
/*********************************/
init_scroll()
{
   int   i;
/* PFSCRL   *sp; */

/* clr_scroll(bg_bank);   */           /* Clear all scroll regs */

   horizon_norm = HORIZON_NORM;        /* Set normal horizon    */
   horizon = bg_bottom = horizon_norm; /* Set initial horizon   */

   bottom_dy =  (SCALE_NORM * Z_LOWEST)/(BOTTOM_LINE - horizon_norm);
   seg_bottom = horizon_norm + ((SCALE_NORM * Z_LOWEST)/bottom_dy);
   slope_norm = PLF_VSIZE/(seg_bottom - horizon_norm);

   bkg_hscroll = SCROLL_CENTER;        /* Set hscroll shadows   */
#if 0
   sp = (PFSCRL *)SCROLL;              /* Set load enable       */
   sp->hscroll = bkg_hscroll | LOAD_ENABLE;   
   sp->vscroll = (((BG_VSIZE - bg_bottom) | LOAD_ENABLE) & 0xFFFC) | bg_bank;
#endif
/*   pf_scroll = TRUE;  NEEDS TO BE REMOVED */

   make_seg();                         /* Draw the playfield    */
   horizon_old = horizon_line;         /* Save current horizon  */
   horizon_ptr = get_scroll(horizon_line); /* Point to horizon line   */
#if 0
   horizon_ptr->vscroll = 0x4000 | LOAD_ENABLE;   /* Set track bottom */
#endif

   make_plf();
/* dump_scroll();   */                 /* Dump the array      */

   pf_scroll = TRUE;
}


/*********************************/
/*   Make segment tables               */
/*********************************/
make_seg()
{
   int   i,slope_idx,new_hidden;
   long   ypos,dy,dscan;
   short   dz,next_strip;
   WORD   prior;
         
   if ((doing_pf) || (pf_scroll == FALSE))
      return;                     /* Don't enter            */

   extra_lines = 0;               /* No extra lines         */
   dz = viewz - height;           /* Get deltaz from playfield      */

   if (height > 0)
      vpoint = (horizon_norm + (height << 2)) & 0xFFC0;   /* Set vanishing point         */
   else
      vpoint = horizon_norm;

   horizon = BOTTOM_LINE;         /* Horizon starts at bottom      */
   trkidx = (viewy + bottom_dy) >> 11;  /* Get track index         */

   if (trkidx >= strip_num)       /* Wrap trkidx if needed */
      trkidx -= strip_num;

   dy = bottom_dy;
   bottomy = ypos = dy + viewy;   /* Get ypos at bottom of first segment      */
   prior = 0xE00;                 /* Start with HIGH priority segments      */
   new_hidden = TRUE;

   vscrl_seg[0] = get_vscroll(get_vnorm(dy));  /* Find normal scan line at bottom of screen   */
   vscan_seg[0] = get_vscan(dy,dz);  /* Get scan line vpos          */

   for (i=1; i < SEG_CNT+1; i++)     /* Loop through segments      */
   {
      next_strip = ypos & (STRIP_SIZE-1);
      ypos += (STRIP_SIZE - next_strip);  /* Next strip's ypos         */
      dy = ypos - viewy;             /* Get strip's deltay from viewpoint   */
      slope_idx = trkidx + i;        /* Get slope index         */
      if (slope_idx >= strip_num) slope_idx -= strip_num;   /* Wrap track slope index */

      if (old_style)
         dz = viewz - (slope_ptr[slope_idx] >> 1); /* Get deltaz from slope table */
      else
         dz = viewz - slope_ptr[slope_idx];        /* Get deltaz from slope table  */
                              
      vscan_seg[i] = get_vscan(dy,dz);             /* Get scan line vpos          */
      vscrl_seg[i] = get_vscroll(get_vnorm(dy));   /* Get vscroll value @ that scanline */

      if (vscan_seg[i] < horizon)            /* If segment < horizon    */
         horizon = vscan_seg[i] & 0xFFC0;    /* Set new horizon         */

      dscan = vscan_seg[i-1] - vscan_seg[i]; /* Get delta scan line count */

      if (dscan < 0)               /* If hidden segment...         */
      {
         if ((new_hidden) && (prior > 0x400))  /* If first hidden segment */
         {
            prior -= 0x400;               /* Bump priority         */
            new_hidden = FALSE;
         }
      }
      else                     /* Here if segment displayed      */
         new_hidden = TRUE;               /* Enable next hidden segment      */

      seg_pri[i-1] = prior;               /* Save priority of this segment   */
   }

   seg_pri[i-1] = prior;               /* Save priority of last segment   */

   bg_bottom = vscan_seg[i-1] - 0x40;         /* Get bottom of background      */

   if (vscan_seg[0] < seg_bottom)         /* If first segment above bottom   */
   {
      lines = vscan_seg[0];            /* Get number of lines to clip      */

      while (lines < seg_bottom)            /* If first segment above bottom   */
      {
         horizon += 0x40;               /* Adjust horizon         */
         vpoint += 0x40;               /* Adjust vanishing point      */
         bg_bottom += 0x40;            /* Adjust bottom of background      */
         lines += 0x40;               /* Count lines            */
      }
   }

   if (abs(bg_bottom - horizon_norm) > 0x800)      /* Limit background V scrolling      */
   {
      if (bg_bottom < horizon_norm)
         bg_bottom = horizon_norm - 0x800;
      else
         bg_bottom = horizon_norm + 0x800;
   }

   horizon_line = (horizon < bg_bottom) ? horizon : bg_bottom;

   if (horizon_line < 0)               /* Set horizon line         */
      horizon_line = 0;
   else
      horizon_line >>= 6;

/* put_vpoint(&vpoint);     */         /* Send vpoint to helper      */

   bckdsp |= B_PLF;            /* Make playfield      */
}


/*********************************/
/*   Make the scrolling playfield  */
/*********************************/
make_plf()
{
   DBL   shift;
   int   tidx,new_hidden;
   short   scrl_frac,lines_save;

   doing_pf = TRUE;          /* Tell others we're doing this      */

#ifdef   DEV
   time_color(0x0F);         /* BLUE               */
#endif

   first_line = TRUE;        /* No lines processed, yet      */
   new_hidden = TRUE;

   scan_pri = 0x18;          /* Start with priority 6 scanlines   */
   curve_acc.full = 0;       /* Clear accumulator         */
   tidx = trkidx;            /* Get track index      */
   linepos = SCROLL_CENTER;  /* First line in center      */
   turn_cur = trkptr[trkidx];  /* Get current turn      */
   max_cacc = 0x100000;
   shift_scale = 0x320;

   stamp_line = SCANLINES-BOTTOM_OFFSET-1-(HORIZON_NORM>>6);

   scan_ptr = &scan_data[SCANLINES-BOTTOM_OFFSET-1];  /* Init scanlines array ptr   */
   scan_ptr->lineh = 0x8003;

   for (seg=0; seg < SEG_CNT; seg++)         /* Loop through strips         */
   {                     /* Get number of lines in this seg   */
      lines = lines_save = (vscan_seg[seg] - vscan_seg[seg+1]) >> 6;   /* Get integer line count */
/*    lut = lutptr[tidx] | lut_mask[tidx];   remove */   /* Get pf LUT            */
      trkang = trkptr[tidx++] << 6;           /* Get track angle         */
      if (tidx >= strip_num) tidx -= strip_num;

      scrl_frac = (vscrl_seg[seg] - vscrl_seg[seg+1]) >> 6; /* Get fraction of displayed picture */
      shift.full = (short)viewx * (short)scrl_frac * shift_scale;   /* Get view shift delta      */

      if (lines > 0)
      {
         ldelta.full = shift.full/lines; 

         if ((first_line == TRUE) &&         /* If no lines have been processed */
          (vscan_seg[seg] >= seg_bottom))    /* And segment still below bottom  */
            extra_lines = (vscan_seg[seg] - seg_bottom) >> 6; /* Skip these lines */

         new_hidden = TRUE;                  /* Enable next hidden segment */
      }
      else
      {
         if (first_line == FALSE)            /* If lines have been processed      */
            xshift.full -= shift.full;       /* Skip the segment         */

         ldelta.full = 0;

         if (new_hidden)                     /* If FIRST hidden segment      */
         {
            if (lines < 0)                   /* If a complete hidden segment      */
            {
               scan_pri -= 4;                /* Change priority on hidden segments   */
               scan_pri &= 0x18;
            }
            else
               scan_pri -= 4;                /* Zero length segment         */

            if (scan_pri < 0) scan_pri = 0;  /* Min check            */
            new_hidden = FALSE;
         }
      }

#ifdef   DEV
      time_color(0x1F);                  /* Full BLUE            */
#endif

      do_segment();                     /* Do this segment         */

#ifdef   DEV
      time_color(0x0F);                  /* Darker BLUE            */
#endif

      vscan_pri[seg] = (scan_pri + 4) << 7;
   }

   vscan_pri[seg] = (scan_pri + 4) << 7;

   while ((--scan_ptr >= scan_data) && (horizon >= bg_bottom))   /* Fill till bottom of background   */
   {
      scan_ptr->linev = 0x4000;               /* Fill value            */
      scan_ptr->lineh = SCROLL_CENTER | scan_pri | lut;
      scan_ptr->lcurve = (SCROLL_CENTER - linepos) << 1;      /* Use last line curve offset      */
      horizon -= 0x40;
   }

   hi_scan = ++scan_ptr;    /* Save highest scanline pointer   */

   bckdsp &= ~B_PLF;        /* Disable dispatch         */
   doing_pf = FALSE;        /* Everything done         */

#ifdef   DEV
   time_color(0);           /* Black            */
#endif

}

/*********************************/
/*   Process one segment            */
/*********************************/
do_segment()
{
   int   lidx;
   WORD   vscroll;

   vscroll = vscrl_seg[seg];      /* Init vscroll line         */

   if (lines > 0)                  /* If we have scan lines      */
   {                              /* Get slope to next segment      */
      slope = (WORD)(vscroll - vscrl_seg[seg+1]) / (WORD)lines;

      while (lines--)               /* Do all lines in this segment      */
      {
         if (extra_lines == 0)            /* If no extra lines..         */
         {
            if (stamp_line--) /*(scan_ptr-- >= scan_data)*//* If in array range */
            {
               if (first_line == TRUE)      /* If this is the first line processed...   */
               {
                  bottom_scroll = vscroll;  /* Save bottom screen vscroll value */
                  /* Get fraction of displayed picture   */
                  line_total = (bottom_scroll - vscrl_seg[SEG_CNT]) >> 6;
                  xshift.full = viewx * line_total * shift_scale;   /* Get view shift delta */
                  first_line = FALSE;

               }

               copy_line((vscroll-0x4000)>>6,stamp_line);
#if 0
               h_scroll[stamp_line] = (linepos + xshift.half.hi) & 0xFFC0;

               /* Add xshift to line position      */
               scan_ptr->lineh = ((linepos + xshift.half.hi) & 0xFFC0) | scan_pri | lut;
               scan_ptr->linev = vscroll & 0xFFC0;         /* Store Vscroll value         */
               scan_ptr->lcurve = (SCROLL_CENTER - linepos) << 1;   /* Save line curve offset */
#endif
            }

            xshift.full -= ldelta.full;         /* Get next view shift         */
         }
         else
            --extra_lines;               /* Skip extra lines         */

         lidx = (BOTTOM_SCRL - vscroll) >> 6;      /* Get scroll line index      */
         curve_val = lacc_tbl[lidx] * trkang;      /* Get new curve value         */
         curve_acc.full += curve_val;         /* Get curve accumulator      */
         linepos += curve_acc.half.hi;         /* Next Hscroll line postion      */

         vscroll -= slope;               /* Next Vscroll pos         */
      }
   }
   else
   {
      if (extra_lines == 0)
         extra_lines = -lines;            /* Calc extra lines in this segment   */
   }
}

